from typing import Callable
from dataclasses import dataclass


@dataclass
class MessageBoxState:
    is_visible: bool = False
    message: str = ""
    ok_button_callback: Callable[[], None] | None = None



from dataclasses import dataclass
from typing import Literal
from rich.text import Text
from textual import on
from textual.app import ComposeResult
from textual.message import Message
from textual.widgets import Tree
from textual.containers import Vertical
from textual.reactive import var

from analyst_klondike.features.code.actions import RunAllCodeAction, RunCodeAndSetResultsAction
from analyst_klondike.features.code_import.ui_actions.actions import ImportTasksAction
from analyst_klondike.state.app_state import INITIAL_STATE, AppState
from analyst_klondike.features.data_context.init_action import InitAction
from analyst_klondike.state.selectors import select_tasks_for_quiz


class Explorer(Vertical):

    DEFAULT_CSS = """
        Explorer {
            border: solid $primary;
            height: 1fr;

            Tree {
                background: $background;

                &:focus {
                    background: $background;
                }
            }
        }
    """

    state = var(INITIAL_STATE, init=False)

    @dataclass
    class ExpNodeData:
        node_type: Literal["root", "group", "quiz", "task"]
        object_id: int | str | None = None

    class TaskSelected(Message):
        """Invokes when task is selected, then display it in editor"""

        def __init__(self, task_id: int | str | None) -> None:
            super().__init__()
            self.task_id = task_id

    class QuizSelected(Message):
        """Invokes when quiz is selected"""

        def __init__(self, quiz_id: str) -> None:
            super().__init__()
            self.quiz_id = quiz_id

    class QuizNodeExpandedOrCollapsed(Message):

        def __init__(self,
                     quiz_id: str,
                     action_type: Literal["expanded", "collapsed"]) -> None:
            super().__init__()
            self.quiz_id = quiz_id
            self.action_type = action_type

    def __init__(self) -> None:
        super().__init__()
        self._is_tree_updated = False

    def on_mount(self) -> None:
        self.border_title = "Навигатор"

    def compose(self) -> ComposeResult:
        yield Tree(self.state.user_email, id="tasks")

    @on(Tree.NodeSelected)
    def on_node_selected(self, event: Tree.NodeSelected[ExpNodeData]) -> None:
        node_data = event.node.data
        if node_data is None:
            return
        if node_data.node_type == "task":
            self.post_message(Explorer.TaskSelected(node_data.object_id))
        elif node_data.node_type == "quiz":
            assert isinstance(node_data.object_id, str)
            self.post_message(Explorer.QuizSelected(node_data.object_id))

    def watch_state(self, new_state: AppState) -> None:
        if not new_state.last_action_type in (InitAction.type,
                                              RunCodeAndSetResultsAction.type,
                                              RunAllCodeAction.type,
                                              ImportTasksAction.type):
            return

        tree: Tree[Explorer.ExpNodeData] = self.query_one(  # type: ignore
            "Tree", Tree)
        tree.clear()
        tree.root.label = Text.from_markup(f":id: {new_state.user_email}")
        tree.root.data = Explorer.ExpNodeData(node_type="root")
        quiz_group_node = tree.root.add(Text.from_markup(":open_file_folder: Тесты"),
                                        expand=True,
                                        data=Explorer.ExpNodeData(node_type="group"))

        for quiz in new_state.data.quizes.values():
            quiz_node = quiz_group_node.add(
                label=Text.from_markup(f":books: {quiz.title}"),
                data=Explorer.ExpNodeData(
                    object_id=quiz.id,
                    node_type="quiz"
                ),
                expand=quiz.is_node_expanded)

            for task in select_tasks_for_quiz(new_state, quiz.id):
                task_node_text = self._get_node_text(new_state, task.id)
                quiz_node.add_leaf(task_node_text,
                                   data=Explorer.ExpNodeData(
                                       object_id=task.id,
                                       node_type="task"))

        tree.root.expand()
        self._is_tree_updated = True

    def _get_node_text(self, state: AppState, task_id: int) -> Text:
        task = state.data.tasks[task_id]
        task_run_result = task.is_passed
        if task_run_result == "not_runned":
            return Text.from_markup(f"{task.title}")
        if task_run_result == "passed":
            return Text.from_markup(f":green_circle: {task.title}")
        if task_run_result == "failed":
            return Text.from_markup(f":red_circle: {task.title}")
        return task_run_result.title

    @on(Tree.NodeExpanded)
    def on_tree_node_expanded(self, ev: Tree.NodeExpanded[ExpNodeData]):
        node_data = ev.node.data
        if node_data is None:
            return
        self._set_expanded_collapsed(node_data, "expanded")

    @on(Tree.NodeCollapsed)
    def on_tree_node_collapsed(self, ev: Tree.NodeCollapsed[ExpNodeData]):
        node_data = ev.node.data
        if node_data is None:
            return
        self._set_expanded_collapsed(node_data, "collapsed")

    def _set_expanded_collapsed(self,
                                node_data: ExpNodeData,
                                action: Literal["expanded", "collapsed"]):
        if node_data.node_type == "quiz" and isinstance(node_data.object_id, str):
            quiz_id = node_data.object_id
            self.post_message(Explorer.QuizNodeExpandedOrCollapsed(
                quiz_id=quiz_id,
                action_type=action
            ))



from analyst_klondike.features.code_import.ui_actions.actions import SwitchToTeacherModeAction
from analyst_klondike.features.events.event_actions import DisableEvents
from analyst_klondike.state.app_state import AppState
from analyst_klondike.state.base_action import BaseAction


def apply(state: AppState, action: BaseAction) -> AppState:
    if isinstance(action, DisableEvents):
        state.event_sender = "no_events"
    if isinstance(action, SwitchToTeacherModeAction):
        state.is_teacher_mode = True
    return state



from typing import Callable
from dataclasses import dataclass
from analyst_klondike.state.base_action import BaseAction


@dataclass
class DisplayMessageBoxAction(BaseAction):
    type = "DISPLAY_MESSAGE_BOX_ACTION"
    message: str
    ok_button_callback: Callable[[], None] | None = None


@dataclass
class HideMessageBoxAction(BaseAction):
    type = "HIDE_MESSAGE_BOX_ACTION"



from abc import ABCMeta, abstractmethod
import httpx


class BaseEventSender(metaclass=ABCMeta):

    @abstractmethod
    async def send_event(self,
                         event_name: str,
                         event_location: str = "/",
                         *,
                         event_json: dict[str, str | int | bool] | None = None) -> None:
        pass

    async def get_user_ip(self) -> str:
        async with httpx.AsyncClient() as client:
            response = await client.get(
                'https://checkip.amazonaws.com/', timeout=5000)
            response.raise_for_status()
            ip_address = response.text.strip()
            return ip_address



from textual.app import ComposeResult
from textual.widgets import Static
from textual.containers import Vertical
from textual.reactive import var

from analyst_klondike.state.app_state import INITIAL_STATE, AppState


class CurrentTaskInfo(Vertical):

    DEFAULT_CSS = """
        CurrentTaskInfo {
            border: solid $primary;
            background: $background;
            height: 1fr;
            padding: 1;
        }
    """

    state = var(INITIAL_STATE, init=False)

    def on_mount(self) -> None:
        self.border_title = "Задача"

    def compose(self) -> ComposeResult:
        yield Static(id="task_text", markup=False)

    def watch_state(self, new_state: AppState) -> None:
        selected_task_id = new_state.current.task_id
        if selected_task_id is not None:
            task = new_state.data.tasks[selected_task_id]
            static = self.query_one("Static", Static)
            static.update(task.description)



from analyst_klondike.features.code_import.ui_actions.actions import ImportTasksAction
from analyst_klondike.features.data_context.data_state import PythonTaskState
from analyst_klondike.state.app_state import AppState
from analyst_klondike.state.base_action import BaseAction


def apply(state: AppState, action: BaseAction) -> AppState:
    if isinstance(action, ImportTasksAction):
        max_task_id = max(t.id for t
                          in state.data.tasks.values())
        task_id = max_task_id + 1
        _remove_with_the_same_title_tasks(action.tasks, state)

        for task in action.tasks:
            task.id = task_id
            state.data.tasks[task_id] = task
            task_id += 1
    return state


def _remove_with_the_same_title_tasks(tasks: list[PythonTaskState], state: AppState):
    imported_tasks_titles = [t.title for t in tasks]
    sorted_dict = {
        k: v for k, v in state.data.tasks.items()
        if v.title not in imported_tasks_titles
    }
    state.data.tasks = sorted_dict



from dataclasses import dataclass
from analyst_klondike.features.data_context.data_state import PythonTaskState
from analyst_klondike.state.base_action import BaseAction


@dataclass
class ImportTasksAction(BaseAction):
    type = "IMPORT_TASKS_ACTION"
    quiz_id: str
    tasks: list[PythonTaskState]


@dataclass
class SwitchToTeacherModeAction(BaseAction):
    type = "SWITCH_TO_TEACHER_MODE"




from dataclasses import dataclass, field
from datetime import datetime
from typing import Any


@dataclass
class ResultCaseState:
    func_params: dict[str, Any]
    expected: str
    actual: str
    passed: bool
    has_exception: bool
    printed_lines: list[str]
    exception_message: str | None


@dataclass
class RunCodeState:
    last_run_dt: datetime = datetime.now()
    last_code: str = ''
    result_cases: list[ResultCaseState] = field(
        default_factory=list["ResultCaseState"]
    )
    has_errors: bool = False
    errors: list[str] = field(default_factory=list[str])



from dataclasses import dataclass
from analyst_klondike.state.base_action import BaseAction


@dataclass
class UpdateCodeAction(BaseAction):
    type = "UPDATE_CODE"
    code: str


@dataclass
class RunCodeAndSetResultsAction(BaseAction):
    type = "RUN_CODE_AND_SET_RESULTS"


@dataclass
class RunAllCodeAction(BaseAction):
    type = "RUN_ALL_CODE_ACTION"
